/*****************************************************************************/
/*!
 * \file expr.cpp
 * 
 * Author: Sergey Berezin
 * 
 * Created: Thu Dec  5 11:35:55 2002
 *
 * <hr>
 * Copyright (C) 2003 by the Board of Trustees of Leland Stanford
 * Junior University and by New York University. 
 *
 * License to use, copy, modify, sell and/or distribute this software
 * and its documentation for any purpose is hereby granted without
 * royalty, subject to the terms and conditions defined in the \ref
 * LICENSE file provided with this distribution.  In particular:
 *
 * - The above copyright notice and this permission notice must appear
 * in all copies of the software and related documentation.
 *
 * - THE SOFTWARE IS PROVIDED "AS-IS", WITHOUT ANY WARRANTIES,
 * EXPRESSED OR IMPLIED.  USE IT AT YOUR OWN RISK.
 * 
 * <hr>
 * 
 */
/*****************************************************************************/


#include <iostream>
#include <sstream>
#include "debug.h"
#include "expr.h"
#include "expr_manager.h"
#include "expr_map.h"
#include "pretty_printer.h"
#include "expr_stream.h"
#include "type.h"
#include "expr_op.h"
#include "notifylist.h"
#include "theorem.h"
#include "theory.h"  // for printing NotifyList
#include <algorithm>

using namespace std;

namespace CVCL {

///////////////////////////////////////////////////////////////////////
// Class Expr methods                                                //
///////////////////////////////////////////////////////////////////////


static Expr recursiveSubst(const Expr& e, const ExprHashMap<Expr>& subst,
			   ExprHashMap<Expr>& visited) {
  // Check the cache.
  // INVARIANT: visited contains all the flagged expressions, and only those
  if(e.getFlag()) return visited[e];

  ExprIndex minIndex = 0;
  for(ExprHashMap<Expr>::iterator i = subst.begin(),iend=subst.end();i!=iend;++i) {
    if(minIndex > (i->first).getIndex())
      minIndex = (i->first).getIndex();
  }

  Expr replaced;

  if(e.isClosure()) {
    const vector<Expr> & vars = e.getVars();
    vector<Expr> common; // Bound vars which occur in subst

    for(vector<Expr>::const_iterator i = vars.begin(), iend=vars.end();
	i!=iend; ++i) {
      if(subst.count(*i) > 0) common.push_back(*i);
    }

    if(common.size() > 0) {
      IF_DEBUG(debugger.counter("substExpr: bound var clashes")++);
      // Reduced substitution (without the common vars)
      ExprHashMap<Expr> newSubst(subst);
      // Remove variables in "common" from the substitution
      for(vector<Expr>::iterator i=common.begin(), iend=common.end();
	  i!=iend; ++i)
	newSubst.erase(*i);

      // Clear all the caches (important!)
      visited.clear();
      e.clearFlags();
      visited = newSubst;

      ExprHashMap<Expr>::iterator j = newSubst.begin();
      for (; j != newSubst.end(); ++j) { // old vars bound in e
	j->first.setFlag();
      }
    
      replaced = 
	newClosureExpr(e.getEM(), e.getKind(), vars,
		       recursiveSubst(e.getBody(), newSubst, visited));

      // Clear the flags again, as we restore the substitution
      visited.clear();
      e.clearFlags();
      visited = subst;
      // Restore the flags on the original substitutions
      for (ExprHashMap<Expr>::iterator i = subst.begin(), iend=subst.end();
	   i != iend; ++i)
	i->first.setFlag();
    } else {
      replaced = newClosureExpr(e.getEM(), e.getKind(), vars,
				recursiveSubst(e.getBody(), subst, visited));
    }
  } else { // Not a Closure
      int changed=0;
      vector<Expr> children;      
      for(Expr::iterator i=e.begin(), iend=e.end(); i!=iend; ++i){	
	Expr repChild = *i;
	if(repChild.getIndex() >= minIndex)
	  repChild = recursiveSubst(*i, subst, visited);
	if(repChild != *i)
	  changed++;
	children.push_back(repChild);
      }
      if(changed > 0)
	replaced = newExpr(getOp(e), children);
      else
	replaced = e;
  }
  visited.insert(e, replaced);
  e.setFlag();
  return replaced;
}


static bool subExprRec(const Expr& e1, const Expr& e2) {
  if(e1 == e2) return true;
  if(e2.getFlag()) return false;
  // e1 is created after e2, so e1 cannot be a subexpr of e2
  if(e1 > e2) return false;
  e2.setFlag();
  bool res(false);
  for(Expr::iterator i=e2.begin(), iend=e2.end(); !res && i!=iend; ++i)
    res = subExprRec(e1, *i);
  return res;
}


bool 
Expr::subExprOf(const Expr& e) const {
  if(*this == e) return true;
  // "this" is created after e, so it cannot be e's subexpression
  if(*this > e) return false;
  clearFlags();
  return subExprRec(*this, e);
}


Expr substExpr(const Expr& e, const vector<Expr>& oldTerms,
		const vector<Expr>& newTerms)
{
  DebugAssert(oldTerms.size() == newTerms.size(), "substExpr: vectors"
	      "don't match in size");
  // Catch the vacuous case
  if(oldTerms.size() == 0) return e;

  ExprHashMap<Expr> oldToNew(10);
  e.clearFlags();
  for(unsigned int i=0; i<oldTerms.size(); i++) {
    oldToNew.insert(oldTerms[i], newTerms[i]);
    oldTerms[i].setFlag();
  }
  // For cache, initialized by the substitution
  ExprHashMap<Expr> visited(oldToNew);
  return recursiveSubst(e, oldToNew, visited);

}


Expr substExpr(const Expr& e, const ExprHashMap<Expr>& oldToNew)
{
  // Catch the vacuous case
  if(oldToNew.size() == 0) return e;
  // Rebuild the map: we'll be using it as a cache
  ExprHashMap<Expr> visited(oldToNew);
  e.clearFlags();
  // Flag all the LHS expressions in oldToNew map.  We'll be checking
  // all flagged expressions (and only those) for substitution.
  for(ExprHashMap<Expr>::iterator i=oldToNew.begin(), iend=oldToNew.end();
      i!=iend; ++i) {
    (*i).first.setFlag();
  }
  return recursiveSubst(e, oldToNew, visited);
}


Expr::Expr(ExprManager *em, const Op& op) {
  vector<Expr> kids;
  d_expr = em->newExprValue(op.getExpr(), kids);
  d_expr->incRefcount();
}

Expr::Expr(ExprManager *em, const Op& op, const Expr& child) {
  DebugAssert(op.getKind() != FORALL && op.getKind() != EXISTS, "");
  std::vector<Expr> kids;
  kids.push_back(child);
  d_expr = em->newExprValue(op.getExpr(), kids);
  d_expr->incRefcount();
}

Expr::Expr(ExprManager *em, const Op& op, const Expr& child0,
	   const Expr& child1) {
  DebugAssert(op.getKind() != FORALL && op.getKind() != EXISTS, "");
  std::vector<Expr> kids;
  kids.push_back(child0);
  kids.push_back(child1);
  d_expr = em->newExprValue(op.getExpr(), kids);
  d_expr->incRefcount();
}

Expr::Expr(ExprManager *em, const Op& op, const Expr& child0,
	   const Expr& child1, const Expr& child2) {
  DebugAssert(op.getKind() != FORALL && op.getKind() != EXISTS, "");
  std::vector<Expr> kids;
  kids.push_back(child0);
  kids.push_back(child1);
  kids.push_back(child2);
  d_expr = em->newExprValue(op.getExpr(), kids);
  d_expr->incRefcount();
}

Expr::Expr(ExprManager *em, const Op& op,
	   const std::vector<Expr>& children) {
  d_expr = em->newExprValue(op.getExpr(), children);
  d_expr->incRefcount();
}

 // These constructors grab the ExprManager from the Op or the first
 // child.  The operator and all children must belong to the same
 // ExprManager.
Expr::Expr(const Op& op) {
  ExprManager* em = op.getEM();
  DebugAssert(em != NULL, "op must have EM");
  vector<Expr> kids;
  d_expr = em->newExprValue(op.getExpr(), kids);
  d_expr->incRefcount();
}

Expr::Expr(const Op& op, const Expr& child) {
  DebugAssert(op.getKind() != FORALL && op.getKind() != EXISTS, "");
  ExprManager* em = op.getEM();
  if(em == NULL) em = child.getEM();
  std::vector<Expr> kids;
  kids.push_back(child);
  d_expr = em->newExprValue(op.getExpr(), kids);
  d_expr->incRefcount();
}
  
Expr::Expr(const Op& op, const Expr& child0, const Expr& child1) {
  DebugAssert(op.getKind() != FORALL && op.getKind() != EXISTS, "");
  ExprManager* em = op.getEM();
  if(em == NULL) em = child0.getEM();
  std::vector<Expr> kids;
  kids.push_back(child0);
  kids.push_back(child1);
  d_expr = em->newExprValue(op.getExpr(), kids);
  d_expr->incRefcount();
}

Expr::Expr(const Op& op, const Expr& child0, const Expr& child1,
	   const Expr& child2) {
  DebugAssert(op.getKind() != FORALL && op.getKind() != EXISTS, "");
  ExprManager* em = op.getEM();
  if(em == NULL) em = child0.getEM();
  std::vector<Expr> kids;
  kids.push_back(child0);
  kids.push_back(child1);
  kids.push_back(child2);
  d_expr = em->newExprValue(op.getExpr(), kids);
  d_expr->incRefcount();
}

Expr::Expr(const Op& op, const std::vector<Expr>& children) {
  ExprManager* em = op.getEM();
  if(em == NULL) {
    DebugAssert(children.size() > 0,
		"Expr::Expr(Op, children): op's EM is NULL and "
		"no children given");
    em = children[0].getEM();
  }
  d_expr = em->newExprValue(op.getExpr(), children);
  d_expr->incRefcount();
}


const Expr& Expr::operator[](int i) const {
  DebugAssert(i < arity(),
	      "CVCL::Expr::operator[i=" + int2string(i)
	      + "]: index >= arity = " + int2string(arity())
	      + "\n  expr = " + toString());
  return (d_expr->getKids())[i];
}

string Expr::toString() const {
  if(isNull()) return "Null";
  ostringstream ss;
  ss << (*this);
  return ss.str();
}

string Expr::toString(InputLanguage lang) const {
  if(isNull()) return "Null";
  ostringstream ss;
  ExprStream os(getEM());
  os.lang(lang);
  os.os(ss);
  os << (*this);
  return ss.str();
}

void Expr::print(InputLanguage lang) const {
  if(isNull()) {
    cout << "Null"; return;
  }
  ExprStream os(getEM());
  os.lang(lang);
  os << *this << endl;
}
  /*! @brief Print only the top node of the expression as AST, and
    submit children to the ExprStream for pretty-printing */
ExprStream& Expr::print(ExprStream& os) const {
  if(isNull()) return os << "Null";
  switch(getKind()) {
  case NULL_KIND: return os << "Null";
  case STRING_EXPR: return os << '"'+ getString() + '"';
  case RATIONAL_EXPR: return os << getRational();
  case SKOLEM_VAR: return os << "SKOLEM_" <<  hash();
  case UCONST: return os << getName();
  case BOOL_VAR: 
    if(isBoolVar())
      return os << "(" << push << "BOOL_VAR" << space << getVar()
		<< push << ")";
    break; // Print as a kind and kids Expr
  case BOUND_VAR: return os << "(BOUND_VAR "+getName()+"_"+getUid()+")";
  case RAW_LIST: {
    os << "(" << push;
    bool firstTime(true);
    for(Expr::iterator i=begin(), iend=end(); i!=iend; ++i) {
      if(firstTime) firstTime = false;
      else os << space;
      os << *i;
    }
    return os << push << ")";
  }
  case FORALL:
  case EXISTS: 
    if(isQuantifier()) {
      os << "(" << push << getEM()->getKindName(getKind())
	 << space << "(" << push;
      const vector<Expr>& vars = getVars();
      vector<Expr>::const_iterator i=vars.begin(), iend=vars.end();
      if(i!=iend) { os << *i; ++i; }
      for(; i!=iend; ++i) os << space << *i;
      os << push << ") " << pop << pop;
      return os << getQuantBody() << push << ")";
    }
    // If not an internal representation of quantifiers, it'll be
    // printed as "normal" Expr with a kind and children
    break;
    /*
  case RECORD_SELECT:
  case RECORD_UPDATE:
    if(isRecordAccess()) {
      os << "(" << push << getEM()->getKindName(getKind()) << space
	 << getRecordField() << space;
      bool firstTime(true);
      for(Expr::iterator i=begin(), iend=end(); i!=iend; ++i) {
	if(firstTime) firstTime = false;
	else os << space;
	os << *i;
      }
      return os << push << ")";
    }
    break;
  case TUPLE_SELECT:
  case TUPLE_UPDATE:
    if(isTupleAccess()) {
      os << "(" << push << getEM()->getKindName(getKind()) << space
	 << getTupleIndex() << space;
      bool firstTime(true);
      for(Expr::iterator i=begin(), iend=end(); i!=iend; ++i) {
	if(firstTime) firstTime = false;
	else os << space;
	os << *i;
      }
      return os << push << ")";
    }
    break;
  case RECORD:
    if(isRecord()) {
      const vector<string>& fields = getRecordFields();
      os << "(RECORD_LITERAL " << push << "(" << push;
      bool firstTime(true);
      for(size_t i=0, iend=fields.size(); i<iend; ++i) {
	if(firstTime) firstTime = false;
	else os << space;
	os << fields[i];
      }
      os << push << ")" << pop << pop << "(" << push;
      firstTime = true;
      for(Expr::iterator i=begin(), iend=end(); i!=iend; ++i) {
	if(firstTime) firstTime = false;
	else os << space;
	os << *i;
      }
      return os << push << ")";
    }
    break;
    */
  default: ; // Don't do anything
  }
  bool isLetDecl(getKind() == LETDECL);
  if(arity()>0) os << "(" << push;
  os << getEM()->getKindName(getKind());
  for(Expr::iterator i=begin(), iend=end(); i!=iend; ++i) {
    if(isLetDecl) os << nodag;
    os << space << *i;
  }
  // Prevent inserting end-of-line before the closing paren
  if(arity()>0) os << push << ")";
  os.resetIndent();
  return os;
}

//! Set initial indentation to n.
/*! The indentation will be reset to default unless the second
    argument is true.  This setting only takes effect when printing to
    std::ostream.  When printing to ExprStream, the indentation can be
    set directly in the ExprStream. */
Expr& Expr::indent(int n, bool permanent) {
  DebugAssert(!isNull(), "Expr::indent called on Null Expr");
  getEM()->indent(n, permanent);
  return *this;
}


void Expr::addToNotify(Theory* i, const Expr& e) const {
  DebugAssert(!isNull(), "Expr::addToNotify() on Null expr");
  if(getNotify() == NULL)
    d_expr->d_notifyList = new NotifyList(getEM()->getCurrentContext());
  getNotify()->add(i, e);
}


void Expr::removeFromNotify(Theory* i, const Expr& e) const {
  DebugAssert(!isNull(), "Expr::removeFromNotify() on Null expr");
  if(getNotify() != NULL)
    getNotify()->remove(i, e);
}


  // This is one of the most friequently called routines.  Make it as
  // efficient as possible.
int compare(const Expr& e1, const Expr& e2) {
  // Quick equality check (operator== is implemented independently
  // and more efficiently)
  if(e1 == e2) return 0;
    
  if(e1.d_expr == NULL) return -1;
  if(e2.d_expr == NULL) return 1;
  // Both are non-Null.  Compare the indices.
  return (e1.getIndex() < e2.getIndex())? -1 : 1;
}

///////////////////////////////////////////////////////////////////////
// Class Expr::iterator methods                                      //
///////////////////////////////////////////////////////////////////////

const Expr& Expr::iterator::operator*() const {
  return *d_it;
}

const Expr* Expr::iterator::operator->() const {
  return &(operator*());
}


Expr::iterator Expr::begin() const {
  if(arity() > 0)
    return iterator(getKids().begin());
  else
    return iterator(getEM()->getEmptyVector().begin());
}


Expr::iterator Expr::end() const {
  if(arity() > 0)
    return iterator(getKids().end());
  else
    return iterator(getEM()->getEmptyVector().end());
}


ostream& operator<<(ostream& os, const Expr& e) {
  if(e.isNull()) return os << "Null";
  ExprStream es(e.getEM());
  es.os(os);
  es << e;
  e.getEM()->restoreIndent();
  return os;
}


ostream& operator<<(ostream& os, const NotifyList& l) {
  os << "NotifyList(\n";
  for(size_t i=0,iend=l.size(); i<iend; ++i) {
    os << "[" << l.getTheory(i)->getName() << ", " << l.getExpr(i) << "]\n";
  }
  return os << ")";
}

// Functions from class Type

Type funType(const std::vector<Type>& typeDom, const Type& typeRan) {
  vector<Expr> tmp;
  for(vector<Type>::const_iterator i=typeDom.begin(), iend=typeDom.end();
      i!=iend; ++i)
    tmp.push_back(i->getExpr());
  tmp.push_back(typeRan.getExpr());
  return Type(newExpr(tmp[0].getEM(), ARROW, tmp));
}


} // end of namespace CVCL
