/*****************************************************************************/
/*!
 * \file expr_op.h
 * \brief Class Op representing the Expr's operator.
 * 
 * Author: Sergey Berezin
 * 
 * Created: Fri Feb  7 15:14:42 2003
 *
 * <hr>
 * Copyright (C) 2003 by the Board of Trustees of Leland Stanford
 * Junior University and by New York University. 
 *
 * License to use, copy, modify, sell and/or distribute this software
 * and its documentation for any purpose is hereby granted without
 * royalty, subject to the terms and conditions defined in the \ref
 * LICENSE file provided with this distribution.  In particular:
 *
 * - The above copyright notice and this permission notice must appear
 * in all copies of the software and related documentation.
 *
 * - THE SOFTWARE IS PROVIDED "AS-IS", WITHOUT ANY WARRANTIES,
 * EXPRESSED OR IMPLIED.  USE IT AT YOUR OWN RISK.
 * 
 * <hr>
 * 
 */
/*****************************************************************************/

// expr.h Has to be included outside of #ifndef, since it sources us
// recursively (read comments in expr_value.h).
#ifndef _CVC_lite__expr_h_
#include "expr.h"
#endif

#ifndef _CVC_lite__expr_op_h_
#define _CVC_lite__expr_op_h_

#include <iostream>
#include <string>
#include "kinds.h"

namespace CVCL {

  class ExprManager;

///////////////////////////////////////////////////////////////////////////////
//                                                                           //
// Class: Op								     //
// Author: Clark Barrett                                                     //
// Created: Wed Nov 27 15:50:38 2002					     //
// Description: Encapsulates all possible Expr operators (including UFUNC)   //
//              and allows switching on the kind.                            //
//              Kinds should be registered with ExprManager.                 //
// 
// Technically, class Op is not part of Expr; it is provided as an
// abstraction for the user.  So, building an Expr from an Op is less
// efficient than building the same Expr directly from the kind.
///////////////////////////////////////////////////////////////////////////////
class Op {
  friend class Expr;
  friend class CInterface;
  // Kind associated with operator.  The kind is either built-in (see
  // kinds.h) or registered using newKind method.
  // int d_kind;

  // Some kinds may require an Expr to store their operator.
  Expr d_expr;
  // ExprManager *d_em;
  

  // Disallow conversion of expr to op
  // Op(const Expr& e);

public:
/////////////////////////////////////////////////////////////////////////
// Public methods
/////////////////////////////////////////////////////////////////////////

  // Construct an operator from a kind.  If ExprManager is provided,
  // Op will be able to print itself nicely.
  // Op(int kind);
  Op(ExprManager *em, int kind);

  //! Constructor for operators with more information than just a kind
  /*! Currently, such operators are function names and quantifiers.
   * For quantifiers, the kind is FORALL or EXISTS, and the expression
   * is VARDECLS with a list of bound variables.
   */
  Op(const Expr& e);


  // Copy constructor
  Op(const Op& op);
  // A constructor that rebuilds the Op for the given ExprManager
  Op(ExprManager* em, const Op& op);
  // Destructor (does nothing)
  ~Op() { }
  // Assignment operator
  Op& operator=(const Op& op);

  // Return the kind of the operator
  int getKind() const { return d_expr.getKind(); }
  // Return the associated ExprManager.  NOTE: it may be NULL.
  ExprManager* getEM() const { return d_expr.getEM(); }

  // Check if op has an associated Expr
  // bool hasExpr() const;


  // Return the expr associated with this operator if applicable.  It
  // is an error to call this when Op has no associated Expr.
  const Expr& getExpr() const {
//     DebugAssert(hasExpr(),
// 		"CVCL::Op::getExpr(" + toString()
// 		 + "): no Expr in this Op");
    return d_expr;
  }

  // Printing functions.

  std::string toString() const;
  friend std::ostream& operator<<(std::ostream& os, const Op& op) {
    return os << "Op(" << op.d_expr << ")";
  }
  // friend bool operator==(const Op& op1, const Op& op2);

}; // end of class Op

  inline Op getOp(const Expr& e) { return Op(e); }

} // end of namespace CVCL

#endif
