/*****************************************************************************/
/*!
 * \file assumptions_value.cpp
 * 
 * Author: Sergey Berezin
 * 
 * Created: Dec 10 00:37:49 GMT 2002
 *
 * <hr>
 * Copyright (C) 2003 by the Board of Trustees of Leland Stanford
 * Junior University and by New York University. 
 *
 * License to use, copy, modify, sell and/or distribute this software
 * and its documentation for any purpose is hereby granted without
 * royalty, subject to the terms and conditions defined in the \ref
 * LICENSE file provided with this distribution.  In particular:
 *
 * - The above copyright notice and this permission notice must appear
 * in all copies of the software and related documentation.
 *
 * - THE SOFTWARE IS PROVIDED "AS-IS", WITHOUT ANY WARRANTIES,
 * EXPRESSED OR IMPLIED.  USE IT AT YOUR OWN RISK.
 * 
 * <hr>
 * 
 */
/*****************************************************************************/
// CLASS: AssumptionsValue
//
// AUTHOR: Sergey Berezin, 12/03/2002
//
// Abstract:
//
// The value holder that class Assumptions points to.
///////////////////////////////////////////////////////////////////////////////

#include <iostream>
#include <sstream>
#include <string>
#include <vector>
#include <iterator>
#include <algorithm>
#include "assumptions_value.h"
#include "assumptions.h"
#include "theorem.h"

namespace CVCL {
  using namespace std;
/////////////////////////////////////////////////////////////////
// class Assumptions methods
/////////////////////////////////////////////////////////////////

// Private constructors for internal use.  Assumes v != NULL.
Assumptions::Assumptions(AssumptionsValue *v)
  : d_val(v) { d_val->d_refcount++; }

  //public constructor
Assumptions::Assumptions(const vector<Theorem>& v) {
  if (v.empty()) {
    d_val = NULL;
    return;
  }

  d_val = new AssumptionsValue(v);
  d_val->d_refcount++;
}


Assumptions::Assumptions(const Theorem& t) {
  vector<Theorem> v(1);
  v[0] = t;
  d_val = new AssumptionsValue(v);
  d_val->d_refcount++;
}


Assumptions::Assumptions(const Theorem& t1, const Theorem& t2) {
  d_val = new AssumptionsValue(t1, t2);
  d_val->d_refcount++;
}


// Destructor
Assumptions::~Assumptions() {
  DebugAssert(d_val == NULL || d_val->d_refcount > 0,
	      "~Assumptions(): refcount = "
	      + int2string(d_val->d_refcount));
  if(d_val != NULL && --(d_val->d_refcount) == 0)
    delete d_val;
}

// Copy constructor.  IMPORTANT: it does NOT create a clean copy
// of the set, but only copies the pointer.  Modifying the copy
// will modify the original value!  Use copy() for cloning.
Assumptions::Assumptions(const Assumptions &assump): d_val(assump.d_val) {
  DebugAssert(d_val == NULL || d_val->d_refcount > 0,
	      "Assumptions(const Assumptions&): refcount = "
	      + int2string(d_val->d_refcount));
  if(d_val != NULL) d_val->d_refcount++;
}

// Assignment.
Assumptions&
Assumptions::operator=(const Assumptions &assump) {
  // Handle self-assignment
  if(this == &assump) return *this;
  if(d_val != NULL) {
    DebugAssert(d_val->d_refcount > 0,
		"Assumptions::operator=: OLD refcount = "
		+ int2string(d_val->d_refcount));		    
    if(--(d_val->d_refcount) == 0) delete d_val;
  }
  d_val = assump.d_val;
  if(d_val != NULL) {
    DebugAssert(d_val->d_refcount > 0,
		"Assumptions::operator=: NEW refcount = "
		+ int2string(d_val->d_refcount));		    
    d_val->d_refcount++;
  }
  return *this;
}
      
void
Assumptions::init() {
  if(isNull()) {
    d_val = new AssumptionsValue;
    d_val->d_refcount++;
  }
}

// Create a clean copy of Assumptions, completely independent of
// the original one
Assumptions
Assumptions::copy() const {
  if(isNull()) return Assumptions();
  // Create a clean copy of the value 
  AssumptionsValue *v = new AssumptionsValue(*d_val);
  return Assumptions(v);
}

// Add a new assumption.  If label is not supplied, generate a new one.
// If we are Null, the initialization happens.
void
Assumptions::add(const Theorem& t) {
  init();
  d_val->add(t);
}

void
Assumptions::clear() {
  DebugAssert(!d_val->d_const, "Can't call clear on const assumptions");
  d_val->d_vector.clear();
}

int
Assumptions::size() const {
  if (isNull())
    return 0;
  else
    return d_val->size();
}

// Return Assumption associated with the expression.  The
// value will be Null if the assumption is not in the set.
//
// NOTE: do not try to assign anything to the result, it won't work.

const Theorem& Assumptions::operator[](const Expr& e) const {
  if (!isNull() && !(d_val->d_vector.empty())) {
    d_val->d_vector.front().clearAllFlags();
  }
  return findTheorem(e);
}

const Theorem& Assumptions::findTheorem(const Expr& e) const {
  static Theorem null;

  if (d_val == NULL) return null;

  const Theorem& t = find(e);
  if (!t.isNull()) return t;
  // recurse
  const vector<Theorem>::iterator aend = d_val->d_vector.end();
  for (vector<Theorem>::iterator iter2 = d_val->d_vector.begin(); 
       iter2 != aend; ++iter2) {
    if (!iter2->isFlagged() && !iter2->isAxiom()) {
      const Theorem& t = iter2->getAssumptions().findTheorem(e);
      if (!t.isNull()) return t;

      iter2->setFlag();
    }
  }
  return null; // not found
}

const Theorem& Assumptions::find(const Expr& e) const {
  static Theorem null;
  if (d_val == NULL) return null;
  return d_val->find(e);
}

////////////////////////////////////////////////////////////////////
// Assumptions::iterator methods
////////////////////////////////////////////////////////////////////

Assumptions::iterator& 
Assumptions::iterator::operator++() { ++d_it; return *this; }

Assumptions::iterator::Proxy 
Assumptions::iterator::operator++(int) { return Proxy(*(d_it++)); }


Assumptions::iterator
Assumptions::begin() const {
  DebugAssert(d_val != NULL,
	      "Thm::Assumptions::begin(): we are Null!");
  return iterator(d_val->d_vector.begin());
}
  

Assumptions::iterator
Assumptions::end() const {
  DebugAssert(d_val != NULL,
	      "Thm::Assumptions::end(): we are Null!");
  return iterator(d_val->d_vector.end()); 
}


// Add all assumptions from the given set
void
Assumptions::add(const Assumptions& a) {
  init();
  d_val->add(*a.d_val);
}

//     size_t size()  const { return (d_val)? d_val->d_map.size() : 0; }
bool
Assumptions::empty() const { return (d_val)? d_val->d_vector.empty() : true ; }

// Check / set the flag indicating that the value is constant
bool
Assumptions::isConst() const { return (d_val)? false : d_val->d_const; }

void
Assumptions::setConst() {
  static AssumptionsValue nullConst(1);
  if(isNull()) {
    d_val = &nullConst;
    d_val->d_refcount++;
  }
  d_val->d_const = true;
}

// int
// Assumptions::scope() const {
//   if(isNull()) return 0;
//   return d_val->scope();
// }

// Merging assumptions
Assumptions
operator+(const Assumptions& a1, const Assumptions& a2) {
  if(a1.isNull()) return a2.copy();
  if(a2.isNull()) return a1.copy();
  return Assumptions((*a1.d_val) + (*a2.d_val));
}

Assumptions
operator-(const Assumptions& a1, const Assumptions& a2) {
  if(a1.isNull() || a2.isNull()) return a1.copy();
  return Assumptions((*a1.d_val) - (*a2.d_val));
}

Assumptions operator-(const Assumptions& a,
		      const Expr& e) {
  if (a.isNull()) return Assumptions();
  if (a.begin() != a.end()) {
    a.begin()->clearAllFlags();
    vector<Theorem> gamma;
    if (findExpr(a, e, gamma)) return Assumptions(gamma);
  } 
  return a.copy();
}

  // recursive helper
bool findExpr(const Assumptions& a, const Expr& e, vector<Theorem>& gamma) { 
  bool found = false;
  const Assumptions::iterator aend = a.end();
  Assumptions::iterator iter = a.begin();
  for (; iter != aend; ++iter) { 
    if (iter->isFlagged()) {
      if (iter->getCachedValue()) found = true;
    }
    else {
      if ((iter->getExpr() == e) || 
	  (!iter->isAxiom() && 
	   findExpr(iter->getAssumptions(), e, gamma))) {
	found = true;
	iter->setCachedValue(true);
      }
      else iter->setCachedValue(false);

      iter->setFlag();
    } 
  }

  if (found) {
    for (iter = a.begin(); iter != aend; ++iter) {     
      if (!iter->getCachedValue()) gamma.push_back(*iter);
    }
  }

  return found;
}

Assumptions operator-(const Assumptions& a,
		      const vector<Expr>& es) {
  if (a.isNull()) return Assumptions();
  if (!es.empty() && a.begin() != a.end()) {
    a.begin()->clearAllFlags();
    vector<Theorem> gamma;
    if (findExprs(a, es, gamma)) return Assumptions(gamma);
  }
  return a.copy();
}

  // recursive helper
bool findExprs(const Assumptions& a, const vector<Expr>& es, 
	       vector<Theorem>& gamma) {
  bool found = false;
  const vector<Expr>::const_iterator esbegin = es.begin();
  const vector<Expr>::const_iterator esend = es.end();
  const Assumptions::iterator aend = a.end();
  Assumptions::iterator iter = a.begin();
  for (; iter != aend; ++iter) {
    if (iter->isFlagged()) {
      if (iter->getCachedValue()) found = true;
    }
    else {
      // switch to binary search below? (sort es first)
      if ((find(esbegin, esend, iter->getExpr()) != esend) ||
	  (!iter->isAxiom() && 
	   findExprs(iter->getAssumptions(), es, gamma))) {
	found = true;
	iter->setCachedValue(true);
      }
      else iter->setCachedValue(false);

      iter->setFlag();
    }
  }

  if (found) {
    for (iter = a.begin(); iter != aend; ++iter) {     
      if (!iter->getCachedValue()) gamma.push_back(*iter);
    }
  }
  
  return found;
}

// Print functions
string
Assumptions::toString() const {
  if(isNull()) return "Null";
  return d_val->toString();
}


void
Assumptions::print() const { cout << toString() << endl; }
      

/////////////////////////////////////////////////////////////////
// class AssumptionsValue methods
/////////////////////////////////////////////////////////////////

  inline bool TheoremEq(const Theorem& t1, const Theorem& t2) 
  { 
    DebugAssert(!t1.isNull() && !t2.isNull(), 
		"AssumptionsValue() Null Theorem passed to constructor");
    return (compare(t1, t2) == 0); 
  }


  // constructor 
  /*!
   * Creates the set of assumptions from a vector of theorems.
   * Non-axiom theorems that do not have premises are not included,
   * since they are generally valid without any context, do not add
   * any information in the assumption tracking.
   */
   AssumptionsValue::AssumptionsValue(const std::vector<Theorem>& v)
      : d_refcount(0), d_const(false) { 
     d_vector.reserve(v.size());
     const vector<Theorem>::const_iterator iend = v.end();
     for (vector<Theorem>::const_iterator i = v.begin(); 
	  i != iend; ++i) {
       if (i->isAxiom() || !i->getAssumptions().empty()) {
	 d_vector.push_back(*i);
       }
     }
     if (d_vector.size() <= 1) return;

     sort(d_vector.begin(), d_vector.end());
     vector<Theorem>::iterator newend =
       unique(d_vector.begin(), d_vector.end(), TheoremEq);
  
     d_vector.resize(newend - d_vector.begin());
   }
  
  // Constructor optimized for a common case 
  AssumptionsValue::AssumptionsValue(const Theorem& t1, const Theorem& t2)
    : d_refcount(0), d_const(false) {
    if (t1.isAxiom() || !t1.getAssumptions().empty()) d_vector.push_back(t1);
    if (t2.isAxiom() || !t2.getAssumptions().empty()) d_vector.push_back(t2);
    if (d_vector.size() <= 1) return;
    
    switch(compare(t1, t2)) {
    case -1: // t1 < t2: correct ordering
      break;
    case 0: // t1 == t2: pop one out
      d_vector.pop_back();
      break;
    case 1: // t1 > t2: swap theorems
      d_vector[0] = t2;
      d_vector[1] = t1;
      break;
    }
  }
  

  // Add a new assumption.  If label is not supplied, generate new one.
  void AssumptionsValue::add(const Theorem& t) {
    DebugAssert(!d_const, 
		"AssumptionsValue::add() called on constant assumptions");
    DebugAssert(!t.isNull(),
		"AssumptionsValue::add() Null Assumption!");

    if (!t.isAxiom() && t.getAssumptions().empty()) return;

    TRACE_MSG("assumptions", "AssumptionsValue::add(" + t.toString() + ")");

    vector<Theorem>::iterator iter = d_vector.begin();
    const vector<Theorem>::const_iterator vend = d_vector.end();
    for (; iter != vend; ++iter) {
      int c(compare(t, *iter));
      if(c == 0) return; // t == *iter
      if(c < 0) break; // t < *iter
    }
    d_vector.insert(iter, t);
  }


    // merge on seperate vector
  void mergeVectors(const vector<Theorem>& v1, const vector<Theorem>& v2,
	       vector<Theorem>& v) {
    v.reserve(v1.size() + v2.size());
    
    vector<Theorem>::const_iterator i = v1.begin();
    vector<Theorem>::const_iterator j = v2.begin();
    const vector<Theorem>::const_iterator v1end = v1.end();
    const vector<Theorem>::const_iterator v2end = v2.end();

    // merge
    while (i != v1end && j != v2end) {
      switch(compare(*i, *j)) {
        case 0:
	  // copy only 1, drop down to next case
	  ++j;
        case -1: // <
	  v.push_back(*i);
	  ++i;
	  break;
        default: // >
	  v.push_back(*j);
	  ++j;
      };
    }
    // Push in the remaining elements
    for(; i != v1end; ++i) v.push_back(*i);
    for(; j != v2end; ++j) v.push_back(*j);
  }



  void AssumptionsValue::add(const AssumptionsValue& a) {
    DebugAssert(!d_const, 
		"AssumptionsValue::add() called on constant assumptions");

    vector<Theorem> v;
    mergeVectors(d_vector, a.d_vector, v);
    d_vector.swap(v);
  }


  const Theorem& AssumptionsValue::find(const Expr& e) const {
    static Theorem null;
    //    binary search
    int lo = 0; 
    int hi = d_vector.size() - 1;
    int loc;
    while (lo <= hi) {
      loc = (lo + hi) / 2;
      
      switch (compare(d_vector[loc], e)) {
        case 0: return d_vector[loc];
        case -1: // t < e
	  lo = loc + 1;
	  break;
        default: // t > e
	  hi = loc - 1;
      };
    }

    return null;
  }

    // Merging assumptions.  NOTE: these genarate new pointers
  AssumptionsValue* operator+(const AssumptionsValue& av1,
			      const AssumptionsValue& av2) {
    AssumptionsValue* a = new AssumptionsValue();
    mergeVectors(av1.d_vector, av2.d_vector, a->d_vector);
    return a;
  }


  AssumptionsValue* operator-(const AssumptionsValue& av1,
			      const AssumptionsValue& av2) {
    AssumptionsValue* a = new AssumptionsValue();
    if (!av1.d_vector.empty()) {
      a->d_vector.reserve(av1.d_vector.size());
      
      vector<Theorem>::const_iterator i = av1.d_vector.begin();
      const vector<Theorem>::const_iterator av1end = av1.d_vector.end();
      const vector<Theorem>::const_iterator av2end = av2.d_vector.end();
      //subtract
      for (vector<Theorem>::const_iterator j = av2.d_vector.begin(); 
	   j != av2end; ++j) {
	while (*i < *j) {
	  a->d_vector.push_back(*i);
	  if (++i == av1end) return a;
	}
	if (compare(*i, *j)==0) { // erase
	  if (++i == av1end) return a;
	}
      }
      // Remaining elements of av1 are larger than av2, push them all in
      for(; i != av1end; ++i) a->d_vector.push_back(*i);
    }
    return a;
  }

  // Print function: print the comma-separated list of assumptions
  string AssumptionsValue::toString() const {
    ostringstream ss;
    ss << (*this);
    return ss.str();
  }

  ostream& operator<<(ostream& os, const Assumptions &assump) {
    if(assump.isNull()) return os << "Null";
    else return os << *assump.d_val;
  }

  // Print assumptions in a nicer way
  ostream &operator<<(ostream& os, const AssumptionsValue &assump) {
    vector<Theorem>::const_iterator i = assump.d_vector.begin();
    const vector<Theorem>::const_iterator iend = assump.d_vector.end();
    if(i != iend) { os << i->getExpr(); i++; }
    for(; i != iend; i++) os << ",\n " << i->getExpr();
    return os;
  }

} // end of namespace CVCL
