/*****************************************************************************/
/*!
 *\file bitvector_expr_value.h
 *\brief Subclasses of ExprValue for bit-vector expressions
 *
 * Author: Sergey Berezin, Vijay Ganesh
 *
 * Created: Wed Jun 23 14:36:59 2004
 *
 * <hr>
 * Copyright (C) 2003 by the Board of Trustees of Leland Stanford
 * Junior University and by New York University. 
 *
 * License to use, copy, modify, sell and/or distribute this software
 * and its documentation for any purpose is hereby granted without
 * royalty, subject to the terms and conditions defined in the \ref
 * LICENSE file provided with this distribution.  In particular:
 *
 * - The above copyright notice and this permission notice must appear
 * in all copies of the software and related documentation.
 *
 * - THE SOFTWARE IS PROVIDED "AS-IS", WITHOUT ANY WARRANTIES,
 * EXPRESSED OR IMPLIED.  USE IT AT YOUR OWN RISK.
 * 
 * <hr>
 * 
 */
/*****************************************************************************/

#ifndef _cvcl__theory_bitvector__bitvector_expr_value_h_
#define _cvcl__theory_bitvector__bitvector_expr_value_h_

#include "theory_bitvector.h"

namespace CVCL {

///////////////////////////////////////////////////////////////////////////////
//class BVConstExpr
///////////////////////////////////////////////////////////////////////////////
//! An expression subclass for bitvector constants.
class BVConstExpr : public ExprValue {
private:
  std::vector<bool> d_bvconst; //!< value of bitvector constant
  size_t d_MMIndex; //!< The registration index for ExprManager
 public:
  //! Constructor
  BVConstExpr(ExprManager* em, std::string bvconst,
	      size_t mmIndex, ExprIndex idx = 0);
  
  //! Constructor
  BVConstExpr(ExprManager* em, std::vector<bool> bvconst,
	      size_t mmIndex, ExprIndex idx = 0);
  
  ExprValue* copy(ExprManager* em, ExprIndex idx = 0) const {
    return new(em->getMM(getMMIndex()))
      BVConstExpr(em, d_bvconst, d_MMIndex, idx);
  }

  size_t computeHash() const;
  size_t getMMIndex() const { return d_MMIndex; }
  size_t isGeneric() const { return getMMIndex(); }
  bool isGeneric(size_t idx) const { return (idx == getMMIndex()); }
  
  //! Only compare the bitvector constant, not the integer attribute
  bool operator==(const ExprValue& ev2) const {
    if(ev2.getMMIndex() != d_MMIndex) return false;
    return (d_bvconst == ((const BVConstExpr&)ev2).d_bvconst);
  }
  
  //! does actual bit extraction
  int getIntValue(int i) const {
    return d_bvconst[i];
  }
  
  //! return the length of the bitvector constant
  size_t getIntValueSize() const {
    return d_bvconst.size();
  }
  
  void* operator new(size_t size, MemoryManager* mm) {
    return mm->newData(size);
  }

  void operator delete(void*) { }
}; //end of BVConstExpr

///////////////////////////////////////////////////////////////////////////////
//class BVPlusExpr
//why we need this class? 
//the number of output bits in a BVPLUS term is a parameter and cannot be 
//construed as child of an EXPR. consequently, we need this expression subclass
//for BVPLUS terms.
///////////////////////////////////////////////////////////////////////////////
//! an expression subclass for bitvector plus terms.
class BVPlusExpr: public ExprNode {
private:
  int d_numOfBits;
  size_t d_MMIndex; //!< The registration index for ExprManager
 public:
  //! Constructor
  BVPlusExpr(ExprManager* em,
	     const std::vector<Expr>& kids, int numOfBits,
	     size_t mmIndex, ExprIndex idx = 0);
  
  ExprValue* copy(ExprManager* em, ExprIndex idx = 0) const {
    return new(em->getMM(getMMIndex()))
      BVPlusExpr(em, d_children, d_numOfBits, d_MMIndex, idx);
  }

  ExprValue* copy(ExprManager* em, const std::vector<Expr>& kids,
		  ExprIndex idx = 0) const {
    return new(em->getMM(getMMIndex()))
      BVPlusExpr(em, kids, d_numOfBits, d_MMIndex, idx);
  }

  int getIntValue(int i) const { return d_numOfBits;}
  size_t getIntValueSize() const { return 1;}
  
  size_t computeHash() const;
  size_t getMMIndex() const { return d_MMIndex; }
  size_t isGeneric() const { return getMMIndex(); }
  bool isGeneric(size_t idx) const { return (idx == getMMIndex()); }

  //! compare the bitvectorplus expr
  bool operator==(const ExprValue& ev2) const {
    if(d_MMIndex != ev2.getMMIndex()) return false;
    if(d_kind != ev2.getKind()) return false;
    if(d_numOfBits != ev2.getIntValue(0)) return false;
    if(d_children != ev2.getKids()) return false;
    return true;
  }
  
  void* operator new(size_t size, MemoryManager* mm) {
    return mm->newData(size);
  }

  void operator delete(void*) { }
}; //end of BVPlusExpr

///////////////////////////////////////////////////////////////////////////////
//class BvParameterExpr
//why we need this class? 
//
//this class is used to construct FIXEDLEFTSHIFT, FIXEDRIGHTSHIFT,
//BOOLEXTRACT, and EXTRACT BVs
//
//LEFTSHIFT: the number of shifts is a parameter
//BOOLEXTRACT: the bit extract position is a parameter
//EXTRACT: the bit extract positions are parameters
//BVMULT: the size of the bitvector is a parameter
//BVLT,BVLE,BVGT,BVGE: the size of the bitvector is the parameter
///////////////////////////////////////////////////////////////////////////////
//! an expression subclass for bitvector plus terms.
class BVParameterExpr: public ExprNode {
private:
  int d_param1;
  int d_param2;
  size_t d_MMIndex; //!< The registration index for ExprManager
 public:
  //! constructor for 2 parameters and takes a vector
  BVParameterExpr(ExprManager* em, const std::vector<Expr>& k,
		  int param1, int param2,
		  size_t mmIndex, int kind, ExprIndex idx = 0);

  //! constructor for 1 parameters and takes a vector
  BVParameterExpr(ExprManager* em, const std::vector<Expr>& k,
		  int param1, size_t mmIndex, 
		  int kind, ExprIndex idx = 0);

  //! Constructor for 2 parameters and 1 expr
  BVParameterExpr(ExprManager* em, const Expr& expr,
		  int param1, int param2, 
		  size_t mmIndex, int kind, ExprIndex idx = 0);
  
  //! Constructor for 1 parameter and 2 expr
  BVParameterExpr(ExprManager* em, const Expr& expr1, 
		  const Expr& expr2, int param,
		  size_t mmIndex, int kind, ExprIndex idx = 0);
    
  //! Constructor for 1 parameter and 1 expr
  BVParameterExpr(ExprManager* em, const Expr& expr,
		  int param,
		  size_t mmIndex, int kind, ExprIndex idx = 0);
  
  //! Constructor for 1 parameter and no children
  BVParameterExpr(ExprManager* em, int param,
		  size_t mmIndex, int kind, ExprIndex idx = 0);
  
  ExprValue* copy(ExprManager* em, ExprIndex idx = 0) const {
    return new(em->getMM(getMMIndex()))
      BVParameterExpr(d_em, d_children, d_param1, d_param2, 
		      d_MMIndex, d_kind, idx);
  }

  ExprValue* copy(ExprManager* em, const std::vector<Expr>& kids,
		  ExprIndex idx = 0) const {
    return new(em->getMM(getMMIndex()))
      BVParameterExpr(d_em, kids, d_param1, d_param2, 
		      d_MMIndex, d_kind, idx);
  }
  int getIntValue(int i) const { 
    if(0 == i)
      return d_param1;
    else
      return d_param2;
  }


  size_t getIntValueSize() const {
    if(EXTRACT == getKind())
      return 2;
    else 
      return 1;
  }
  
  size_t computeHash() const;
  size_t getMMIndex() const { return d_MMIndex; }
  size_t isGeneric() const { return getMMIndex(); }
  bool isGeneric(size_t idx) const { return (idx == getMMIndex()); }
  
  //! Only compare the bitvector constant, not the integer attribute
  bool operator==(const ExprValue& ev2) const {
    if(ev2.getMMIndex() != d_MMIndex) return false;
    if(getKind() != ev2.getKind()) return false;
    if(d_param1 != ev2.getIntValue(0)) return false;
    if(d_param2 != ev2.getIntValue(1)) return false;
    if(d_children != ev2.getKids()) return false;
    return true;
  }
  
  void* operator new(size_t size, MemoryManager* mm) {
    return mm->newData(size);
  }

  void operator delete(void*) { }
}; //end of BVParameterExpr

/*****************************************************************************/
/*!
 *\class BVTypePredExpr
 *\brief BVTypePredExpr
 *
 * Author: Sergey Berezin
 *
 * Created: Tue Jan 11 14:36:58 2005
 *
 * 
 */
/*****************************************************************************/
class BVTypePredExpr : public ExprValue {
private:
  Expr d_type;
  Expr d_expr;
  size_t d_MMIndex; //!< The registration index for ExprManager
 public:
  //! Constructor
  BVTypePredExpr(ExprManager* em, const Type& t, const Expr& e,
		 size_t mmIndex, ExprIndex idx = 0)
    : ExprValue(em, BVTYPEPRED, idx), d_type(t.getExpr()), d_expr(e),
      d_MMIndex(mmIndex) { }
  
  ExprValue* copy(ExprManager* em, ExprIndex idx = 0) const {
    return new(em->getMM(getMMIndex()))
      BVTypePredExpr(em, d_type, d_expr, d_MMIndex, idx);
  }

  ExprValue* rebuild(ExprManager* em) const {
    return new(em->getMM(getMMIndex()))
      BVTypePredExpr(em, rebuildExpr(em, d_type),
		     rebuildExpr(em, d_expr), d_MMIndex);
  }

  size_t computeHash() const;
  size_t getMMIndex() const { return d_MMIndex; }
  size_t isGeneric() const { return getMMIndex(); }
  bool isGeneric(size_t idx) const { return (idx == getMMIndex()); }
  
  //! Only compare the bitvector constant, not the integer attribute
  bool operator==(const ExprValue& ev2) const {
    if(ev2.getMMIndex() != d_MMIndex) return false;
    return (d_kind == ev2.getKind()
	    && d_type == ev2.getExprValue(0)
	    && d_expr == ev2.getExprValue(1));
  }

  const Expr& getExprValue(int idx) const {
    DebugAssert(0<=idx && idx<=1,
		"BVTypePredExpr::getExprValue: idx is out of range");
    if(idx==0) return d_type;
    else return d_expr;
  }

  size_t getExprValueSize() const { return 2; }

  
  void* operator new(size_t size, MemoryManager* mm) {
    return mm->newData(size);
  }

  void operator delete(void*) { }
}; //end of BVConstExpr

} // end of namespace CVCL
#endif
